# -*- coding: utf-8 -*-
"""King chess piece."""

from .figure import Figure, FieldMustBeCastledError, IllegalMoveError


class King(Figure):

    """King chess piece."""

    start_position = {"white": ["E1"],
                      "black": ["E8"]}

    def __str__(self):
        if self.color == "white":
            string = "♔"
        else:
            string = "♚"
        return string

    def __init__(self, player):
        """Initialize king and set castle positions."""
        super(King, self).__init__(player)
        self.castle_positions = {"left": self.position.to_left().to_left(),
                                 "right": self.position.to_right().to_right()}

    @property
    def legal_moves(self):
        """Return legal moves from current position."""
        moves = []
        # Castling positions
        if not self.already_moved:
            moves.extend(self.castle_positions.values())
        # Normal positions
        moves.extend(self._fields_in_directions(("to_left", "to_right",
                                                 "above", "below",
                                                 "above_right", "below_right",
                                                 "above_left", "below_left"),
                                                1))
        return moves

    def move(self, field):
        """
        Move to field.

        If given field is a castling field and castling is legal, use try
        castling. Otherwise call superclass method.
        """
        if (field in self.legal_moves and
                field in self.castle_positions.values()):
            raise FieldMustBeCastledError("Goal field is for castling.")
        else:
            super(King, self).move(field)

    @property
    def in_check(self):
        """Test if king is in check."""
        return self.position.threatened_by(self.player.opponent)

    def castle(self, field):
        """
        Castle king.

        First, check if corresponding rook is there and has not moved yet. Then
        check if any of the positions between and including the king and rook
        are occupied or threatened. If this is not the case, castle the king.
        If not, raise error.
        """
        involved_fields = []
        rook, rook_position = None, None
        position = old_position = self.position
        involved_fields.append(position)

        if field == self.castle_positions["left"]:
            rook_position = self.player.chessboard.fields[
                                                      "A{}".format(
                                                      self.position.number)]
            while position != self.castle_positions["left"]:
                position = position.to_left()
                involved_fields.append(position)
        # Rook is on right
        elif field == self.castle_positions["right"]:
            rook_position = self.player.chessboard.fields[
                                                      "H{}".format(
                                                      self.position.number)]
            while position != self.castle_positions["right"]:
                position = position.to_right()
                involved_fields.append(position)

        if rook_position:
            rook = rook_position.figure
        if rook and not rook.already_moved:
            involved_fields_free = True
            for position in involved_fields:
                if position.threatened_by(self.player.opponent):
                    involved_fields_free = False
                elif position.figure and not (position == self.position or
                                              position == rook_position):
                    involved_fields_free = False
            if involved_fields_free:
                field.receive_figure(self)
                if field == self.castle_positions["left"]:
                    field.to_right().receive_figure(rook)
                else:
                    field.to_left().receive_figure(rook)

        # King could not be castled
        if self.position != field:
            raise IllegalMoveError("King cannot castle here.")
        else:
            return rook



#!/bin/env python
# -*- coding: utf-8 -*-

"""The GUI for playing chess using KingSnake!"""

import pickle
import sys
import time

from king_snake.player import Player
from king_snake.chessboard import Chessboard
from king_snake.errors import ChessError


class ChessGame(object):

    """A chess game manager"""

    def __init__(self):
        self.white, self.black = Player(), Player()
        self.chessboard = Chessboard()
        self.chessboard.add_players(self.white, self.black)
        self.greet()

    def greet(self):
        """Greet the players and begin game."""
        print("Welcome to KingSnake.\n"
              "Enter moves directly or press Enter to enter the menu.")
        self.get_move()

    def get_move(self, message=""):
        """Show board and prompt for current move, then execute it."""
        valid_move = False
        while not valid_move:
            self.show(message)
            move = raw_input("Please enter your move (e.g. E2 E4) or enter to "
                             "access the menu: ")
            if not move:
                self.menu()
            else:
                try:
                    start_position, end_position = move.split()
                    self.chessboard.current_player.move(start_position.upper(),
                                                        end_position.upper())
                    valid_move = True
                except ValueError:
                    message = ("Please enter your move in the following form:"
                               "\n{start position} {end position}\n"
                              "Positions are notated using their letter "
                              "followed by their number.\n"
                              "Example valid move request to move from A1 to "
                              "A2: 'A1 A2'")
                except KeyError:
                    message = "Only valid fields are allowed."
                except ChessError as error:
                    message = error
        self.get_move(message)

    def menu(self):
        """Allow user to do something other than move pieces"""

        def quit_game():
            """Quit game."""
            sys.exit()

        def resign():
            """Resign."""
            print("{} resigns. {} is the winner!".format(
                 self.chessboard.current_player.color.capitalize(),
                 self.chessboard.current_player.opponent.color.capitalize()))
            time.sleep(10)
            quit_game()

        def restart():
            """Start new game."""
            self.__init__()

        def save_game():
            """Save game to file."""
            file_name = raw_input("What file would you like to save to?: ")
            try:
                with open(file_name, "w") as saved_game:
                    pickle.dump(self.chessboard, saved_game)
            except IOError:
                self.get_move("The file you have chosen is invalid. "
                              "Please enter a valid filename.")
            self.get_move("Game saved to {}.".format(file_name))

        def load_game():
            """Load game from file."""
            file_name = raw_input("What file would you like to load from?: ")
            try:
                with open(file_name) as saved_game:
                    self.chessboard = pickle.load(saved_game)
                    self.white = self.chessboard.players["white"]
                    self.black = self.chessboard.players["black"]
            except IOError:
                self.get_move("The file you have chosen is invalid. "
                              "Please enter a valid filename.")

        def undo_turn():
            """Undo turn."""
            self.chessboard.rollback()
            self.get_move("Move restored.")

        def return_to_game():
            """Resume play."""
            self.get_move()

        menu_choices = []
        for function in (quit_game, resign, restart, save_game, load_game,
                         undo_turn, return_to_game):
            menu_choices.append((function, function.__doc__))
        valid_choice = False
        while not valid_choice:
            for number, choice in enumerate(menu_choices):
                print("{}. {}".format(number + 1, choice[1]))
            try:
                decision = int(raw_input("What would you like to do?: ")) - 1
                try:
                    valid_choice = True
                    menu_choices[decision][0]()
                except IndexError:
                    valid_choice = False
                    print("Please enter a valid menu number.")
            except (ValueError, IndexError):
                print("Please enter a valid menu number.")

    def show(self, message=""):
        """Show chessboard and print current player."""
        print(self.chessboard)
        print("{}\n"
              "It's {}'s turn.".format(message,
                                       self.chessboard.current_player.color))

game = ChessGame()



# -*- coding: utf-8 -*-
"""A chess board and fields."""

from king_snake.errors import FieldOccupiedError


class Field(object):

    """A field on a chessboard."""

    def __repr__(self):
        return "Field({letter}, {number}, {board})".format(
                                           letter=chr(self.letter),
                                           number=self.number,
                                           board=self.chessboard.__repr__())

    def __str__(self):
        """
        Return unicode representation of the field and its figure, if present.

        If no figure is present, "..." is returned.
        """
        figure = self.figure
        if not figure:
            figure = "…"
        return str(figure)

    def __init__(self, letter, number, chessboard):
        """
        Store position on chessboard.

        Letter is given as a capital character and converted internally into an
        integer.
        """
        self.letter = ord(letter)
        self.number = number
        self.chessboard = chessboard
        self.figure = None

    def receive_figure(self, figure):
        """
        Place a figure on field.

        If figure was already on chessboard, remove figure from former field.
        Set figure's position to self.
        If self is occupied, raise error.
        """
        if not self.figure:
            self.figure = figure
            if figure.position:
                figure.position.figure = None
            figure.position = self
        else:
            raise FieldOccupiedError("The goal field is occupied.")

    def to_right(self):
        """Return field to right of self or None if there is none."""
        position = chr(self.letter + 1) + str(self.number)
        return self.chessboard.fields.get(position)

    def to_left(self):
        """Return field to left of self or None if there is none."""
        position = chr(self.letter - 1) + str(self.number)
        return self.chessboard.fields.get(position)

    def above(self):
        """Return field above self or None if there is none."""
        position = chr(self.letter) + str(self.number + 1)
        return self.chessboard.fields.get(position)

    def below(self):
        """Return field below self or None if there is none."""
        position = chr(self.letter) + str(self.number - 1)
        return self.chessboard.fields.get(position)

    def above_right(self):
        """Return field above self and to right or None if there is none."""
        above = self.above()
        if above:
            return above.to_right()

    def above_left(self):
        """Return field above self and to left or None if there is none."""
        above = self.above()
        if above:
            return above.to_left()

    def below_right(self):
        """Return field below self and to right or None if there is none."""
        below = self.below()
        if below:
            return below.to_right()

    def below_left(self):
        """Return field below self and to left or None if there is none."""
        below = self.below()
        if below:
            return below.to_left()

    def threatened_by(self, player):
        """Field is in legal moves of given player."""
        for figure in player.figures:
            if figure.position and self in figure.legal_moves:
                return True


class Chessboard(object):

    """A chessboard."""

    def __str__(self):
        """Print border with coordinates and all fields in order."""
        string = ""
        column_string = ""
        if self.current_player.color == "white":
            row_range = (8, 0, -1)
            column_range = (ord("A"), ord("H") + 1)
        else:
            row_range = (1, 9)
            column_range = (ord("H"), ord("A") - 1, -1)
        for row in range(*row_range):
            string += "{}|".format(row)
            for column in range(*column_range):
                position = "{}{}".format(chr(column), row)
                string += str(self.fields[position])
                if len(column_string) < 8:
                    column_string += chr(column)
            string += "\n"
        string += " " + "=" * 9 + "\n"
        string += "  {}".format(column_string)
        return string

    def __init__(self, players=None, move_history=None):
        """
        Initialize fields and set current move to 1.

        Players do not have to be supplied. If they are, they should be
        provided as a dictionary: {"white": Player, "black": Player}

        If a move history is supplied, apply that move history.
        """
        self.fields = dict()
        for letter in "ABCDEFGH":
            for number in range(1, 9):
                self.fields[letter + str(number)] = Field(letter, number, self)

        if players:
            self.add_players(players["white"], players["black"])
        else:
            self.players = dict()
            self.current_player = None

        self.move_history = []
        self.current_move = 1
        if move_history:
            for start_field, goal_field in move_history:
                self.current_player.move(start_field, goal_field)

    def rollback(self):
        """Rollback to previous state"""
        self.__init__(self.players, self.move_history[:-1])

    def add_players(self, white, black):
        """Add players to the game, assign colors and set up board."""
        for color, player in (("white", white), ("black", black)):
            self.players[color] = player
            self.players[color].color = color
            player.set_up_board(self)
            self.current_player = self.players["white"]

    def end_turn(self, start_field, goal_field):
        """
        End turn for current player and increment current_move.

        The move history is used for rollback capabilities
        """
        self.current_move += 1
        self.move_history.append((start_field, goal_field))
        if self.current_player == self.players["white"]:
            self.current_player = self.players["black"]
        else:
            self.current_player = self.players["white"]
